/*
    $Id: bitsobj.c 686 2014-04-05 06:17:43Z soci $

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <string.h>
#include <math.h>
#include "values.h"
#include "bitsobj.h"
#include "eval.h"
#include "unicode.h"
#include "encoding.h"
#include "boolobj.h"

static struct obj_s obj;

obj_t BITS_OBJ = &obj;

static void destroy(struct value_s *v1) {
    if (v1->u.bits.val != v1->u.bits.data) free(v1->u.bits.data);
}

static bdigit_t *bnew(struct value_s *v, size_t len) {
    if (len > 2) {
        bdigit_t *s = (bdigit_t *)malloc(len * sizeof(bdigit_t));
        if (!s || len > SIZE_MAX / sizeof(bdigit_t)) err_msg_out_of_memory(); /* overflow */
        return s; 
    }
    return v->u.bits.val;
}

static void copy(const struct value_s *v1, struct value_s *v) {
    v->obj = v1->obj;
    v->u.bits.bits = v1->u.bits.bits;
    v->u.bits.inv = v1->u.bits.inv;
    v->u.bits.len = v1->u.bits.len;
    if (v1->u.bits.len) {
        v->u.bits.data = bnew(v, v->u.bits.len);
        memcpy(v->u.bits.data, v1->u.bits.data, v->u.bits.len * sizeof(bdigit_t));
    } else {
        v->u.bits.val[0] = 0;
        v->u.bits.data = v->u.bits.val;
    }
}

static void copy_temp(const struct value_s *v1, struct value_s *v) {
    v->obj = BITS_OBJ;
    v->u.bits.bits = v1->u.bits.bits;
    v->u.bits.inv = v1->u.bits.inv;
    v->u.bits.len = v1->u.bits.len;
    if (v1->u.bits.len) {
        if (v1->u.bits.data == v1->u.bits.val) {
            v->u.bits.data = v->u.bits.val;
            memcpy(v->u.bits.data, v1->u.bits.data, v->u.bits.len * sizeof(bdigit_t));
        } else v->u.bits.data = v1->u.bits.data;
    } else {
        v->u.bits.val[0] = 0;
        v->u.bits.data = v->u.bits.val;
    }
}

static int same(const struct value_s *v1, const struct value_s *v2) {
    if (v2->obj != BITS_OBJ || v1->u.bits.len != v2->u.bits.len || v1->u.bits.bits != v2->u.bits.bits || v1->u.bits.inv != v2->u.bits.inv) return 0;
    return !memcmp(v1->u.bits.data, v2->u.bits.data, v1->u.bits.len * sizeof(bdigit_t));
}

static int truth(const struct value_s *v1, struct value_s *v, enum truth_e type, linepos_t epoint) {
    int result;
    size_t i;
    bdigit_t b;
    const char *name;
    switch (type) {
    case TRUTH_ALL:
        result = 1;
        for (i = 0; i < v1->u.bits.bits / 8 / sizeof(bdigit_t); i++) {
            b = (i >= v1->u.bits.len) ? 0 : v1->u.bits.data[i];
            if (!v1->u.bits.inv) b = ~b;
            if (b) {result = 0; goto end;}
        }
        b = (i >= v1->u.bits.len) ? 0 : v1->u.bits.data[i];
        if (!v1->u.bits.inv) b = ~b;
        b <<= 8 * sizeof(bdigit_t) - v1->u.bits.bits % (8 * sizeof(bdigit_t));
        if (b) result = 0;
        break;
    case TRUTH_ANY:
        result = 0;
        for (i = 0; i < v1->u.bits.bits / 8 / sizeof(bdigit_t); i++) {
            b = (i >= v1->u.bits.len) ? 0 : v1->u.bits.data[i];
            if (v1->u.bits.inv) b = ~b;
            if (b) {result = 1; goto end;}
        }
        b = (i >= v1->u.bits.len) ? 0 : v1->u.bits.data[i];
        if (v1->u.bits.inv) b = ~b;
        b <<= 8 * sizeof(bdigit_t) - v1->u.bits.bits % (8 * sizeof(bdigit_t));
        if (b) result = 1;
        break;
    case TRUTH_BOOL:
        result = !!v1->u.bits.len || v1->u.bits.inv;
        break;
    default:
        name = v1->obj->name;
        if (v1 == v) destroy(v);
        v->obj = ERROR_OBJ;
        v->u.error.num = ERROR_____CANT_BOOL;
        v->u.error.epoint = *epoint;
        v->u.error.u.objname = name;
        return 1;
    }
end:
    if (v1 == v) destroy(v);
    bool_from_int(v, result);
    return 0;
}

static void repr(const struct value_s *v1, struct value_s *v, linepos_t UNUSED(epoint)) {
    size_t len, i, len2;
    uint8_t *s;
    struct value_s tmp;

    len2 = v1->u.bits.bits;
    if (len2 & 3) {
        len = 1 + v1->u.bits.inv;
        len += len2;
        s = str_create_elements(&tmp, len);

        len = 0;
        if (v1->u.bits.inv) s[len++] = '~';
        s[len++] = '%';
        for (i = len2; i--;) {
            size_t j = i / 8 / sizeof(bdigit_t);
            s[len++] = (j >= v1->u.bits.len) ? 0x30 : 0x30 | ((v1->u.bits.data[j] >> (i & (8 * sizeof(bdigit_t) - 1))) & 1);
        }
    } else {
        static const char *hex = "0123456789abcdef";
        len2 /= 4;
        len = 1 + v1->u.bits.inv;
        len += len2;
        s = str_create_elements(&tmp, len);

        len = 0;
        if (v1->u.bits.inv) s[len++] = '~';
        s[len++] = '$';
        for (i = len2; i--;) {
            size_t j = i / 2 / sizeof(bdigit_t);
            s[len++] = (j >= v1->u.bits.len) ? 0x30 : (uint8_t)hex[(v1->u.bits.data[j] >> ((i & (2 * sizeof(bdigit_t) - 1)) * 4)) & 15];
        }
    }
    if (v == v1) v->obj->destroy(v);
    v->obj = STR_OBJ;
    if (s == tmp.u.str.val) {
        memcpy(v->u.str.val, s, len);
        s = v->u.str.val;
    }
    v->u.str.len = len;
    v->u.str.chars = len;
    v->u.str.data = s;
}

static int hash(const struct value_s *v1, struct value_s *v, linepos_t epoint) {
    struct value_s tmp;
    int ret;

    switch (v1->u.bits.len) {
    case 0: return (-v1->u.bits.inv) & ((~(unsigned int)0) >> 1);
    case 1: return (v1->u.bits.inv ? ~v1->u.bits.data[0] : v1->u.bits.data[0]) & ((~(unsigned int)0) >> 1);
    case 2: return (v1->u.bits.inv ? ~(v1->u.bits.data[0] | (v1->u.bits.data[1] << 16)) : (v1->u.bits.data[0] | (v1->u.bits.data[1] << 16))) & ((~(unsigned int)0) >> 1);
    }
    int_from_bits(&tmp, v1);
    ret = tmp.obj->hash(&tmp, v, epoint);
    tmp.obj->destroy(&tmp);
    return ret;
}

static int MUST_CHECK ival(const struct value_s *v1, struct value_s *v, ival_t *iv, int bits, linepos_t epoint) {
    struct value_s tmp, tmp2;
    int ret;
    int_from_bits(&tmp, v1);
    ret = tmp.obj->ival(&tmp, &tmp2, iv, bits, epoint);
    tmp.obj->destroy(&tmp);
    if (ret) {
        if (v1 == v) destroy(v);
        tmp2.obj->copy_temp(&tmp2, v);
    }
    return ret;
}

static int MUST_CHECK uval(const struct value_s *v1, struct value_s *v, uval_t *uv, int bits, linepos_t epoint) {
    switch (v1->u.bits.len) {
    case 2: *uv = v1->u.bits.data[0] | (v1->u.bits.data[1] << (8 * sizeof(bdigit_t)));
            if (bits < 2 * 8 * (int)sizeof(bdigit_t) && *uv >> bits) break;
            if (v1->u.bits.inv) *uv = ~*uv;
            return 0;
    case 1: *uv = v1->u.bits.data[0];
            if (bits < 8 * (int)sizeof(bdigit_t) && *uv >> bits) break;
            if (v1->u.bits.inv) *uv = ~*uv;
            return 0;
    case 0: *uv = v1->u.bits.inv ? ~(uval_t)0 : 0; return 0;
    default: break;
    }
    *uv = 0;
    if (v1 == v) destroy(v);
    v->obj = ERROR_OBJ;
    v->u.error.num = ERROR_____CANT_UVAL;
    v->u.error.u.bits = bits;
    v->u.error.epoint = *epoint;
    return 1;
}

static int MUST_CHECK real(const struct value_s *v1, struct value_s *v, double *r, linepos_t epoint) {
    size_t i, len1 = v1->u.bits.len;
    double d = -v1->u.bits.inv;
    for (i = 0; i < len1; i++) {
        if (v1->u.bits.inv) d -= ldexp((double)v1->u.bits.data[i], i * 8 * sizeof(bdigit_t));
        else d += ldexp((double)v1->u.bits.data[i], i * 8 * sizeof(bdigit_t));
        if (d == HUGE_VAL) {
            const char *name = v1->obj->name;
            if (v1 == v) destroy(v);
            v->obj = ERROR_OBJ;
            v->u.error.num = ERROR_____CANT_REAL;
            v->u.error.epoint = *epoint;
            v->u.error.u.objname = name;
            return 1;
        }
    }
    *r = d;
    return 0;
}

static void sign(const struct value_s *v1, struct value_s *v, linepos_t UNUSED(epoint)) {
    int s;
    if (!v1->u.bits.len) s = 0;
    else if (v1->u.bits.inv) s = -1;
    else s = 1;
    if (v1 == v) destroy(v);
    int_from_int(v, s);
}

static void absolute(const struct value_s *v1, struct value_s *v, linepos_t epoint) {
    struct value_s tmp;
    int_from_bits(&tmp, v1);
    tmp.obj->abs(&tmp, v, epoint);
    tmp.obj->destroy(&tmp);
}

static void integer(const struct value_s *v1, struct value_s *v, linepos_t UNUSED(epoint)) {
    struct value_s tmp;
    int_from_bits(&tmp, v1);
    if (v == v1) v->obj->destroy(v);
    tmp.obj->copy_temp(&tmp, v);
}

static void len(const struct value_s *v1, struct value_s *v, linepos_t UNUSED(epoint)) {
    size_t uv = v1->u.bits.bits;
    if (v == v1) destroy(v);
    int_from_uval(v, uv);
}

static void bits_from_bool(struct value_s *v, int i) {
    v->obj = BITS_OBJ;
    v->u.bits.data = v->u.bits.val;
    v->u.bits.val[0] = i;
    v->u.bits.len = (i != 0);
    v->u.bits.inv = 0;
    v->u.bits.bits = 1;
}

void bits_from_bools(struct value_s *v, int i, int j) {
    v->obj = BITS_OBJ;
    v->u.bits.data = v->u.bits.val;
    v->u.bits.val[0] = (i << 1) | j;
    v->u.bits.len = (v->u.bits.val[0] != 0);
    v->u.bits.inv = 0;
    v->u.bits.bits = 2;
}

void bits_from_u8(struct value_s *v, uint8_t i) {
    v->obj = BITS_OBJ;
    v->u.bits.data = v->u.bits.val;
    v->u.bits.val[0] = i;
    v->u.bits.len = (i != 0);
    v->u.bits.inv = 0;
    v->u.bits.bits = 8;
}

void bits_from_u16(struct value_s *v, uint16_t i) {
    v->obj = BITS_OBJ;
    v->u.bits.data = v->u.bits.val;
    v->u.bits.val[0] = i;
    v->u.bits.len = (i != 0);
    v->u.bits.inv = 0;
    v->u.bits.bits = 16;
}

static void bits_from_u24(struct value_s *v, int i) {
    v->obj = BITS_OBJ;
    v->u.bits.data = v->u.bits.val;
    v->u.bits.val[0] = i;
    v->u.bits.val[1] = i >> 16;
    v->u.bits.len = (i != 0);
    v->u.bits.inv = 0;
    v->u.bits.bits = 24;
}

size_t bits_from_hexstr(struct value_s *v, const uint8_t *s) {
    size_t i = 0, j = 0, sz, l;
    bdigit_t uv = 0;
    int bits = 0;
    bdigit_t *d;
    v->obj = BITS_OBJ;

    while ((s[i] ^ 0x30) < 10 || (uint8_t)((s[i] | 0x20) - 0x61) < 6) i++;
    if (!i) {
        v->u.bits.data = v->u.bits.val;
        v->u.bits.val[0] = 0;
        v->u.bits.len = 0;
        v->u.bits.inv = 0;
        v->u.bits.bits = 0;
        return 0;
    }
    l = i;
    sz = i / 2 / sizeof(bdigit_t);
    if (i % (2 * sizeof(bdigit_t))) sz++;

    d = bnew(v, sz);

    while (i) {
        i--;
        if (s[i] < 0x40) uv |= (s[i] & 15) << bits;
        else uv |= ((s[i] & 7) + 9) << bits;
        if (bits == (8 * sizeof(bdigit_t) - 4)) {
            d[j++] = uv;
            bits = uv = 0;
        } else bits += 4;
    }
    if (bits) d[j] = uv;

    while (sz && !d[sz - 1]) sz--;
    if (sz <= 2 && v->u.bits.val != d) {
        memcpy(v->u.bits.val, d, sz);
        free(d);
        d = v->u.bits.val;
    }
    v->u.bits.data = d;
    v->u.bits.len = sz;
    v->u.bits.inv = 0;
    v->u.bits.bits = j * sizeof(bdigit_t) * 8 + bits;
    return l;
}

size_t bits_from_binstr(struct value_s *v, const uint8_t *s) {
    size_t i = 0, j = 0, sz, l;
    uval_t uv = 0;
    int bits = 0;
    bdigit_t *d;
    v->obj = BITS_OBJ;

    while ((s[i] & 0xfe) == 0x30) i++;
    if (!i) {
        v->u.bits.data = v->u.bits.val;
        v->u.bits.val[0] = 0;
        v->u.bits.len = 0;
        v->u.bits.inv = 0;
        v->u.bits.bits = 0;
        return 0;
    }
    l = i;
    sz = i / 8 / sizeof(bdigit_t);
    if (i % (8 * sizeof(bdigit_t))) sz++;

    d = bnew(v, sz);

    while (i) {
        i--;
        if (s[i] == 0x31) uv |= 1 << bits;
        if (bits == (8 * sizeof(bdigit_t) - 1)) {
            d[j++] = uv;
            bits = uv = 0;
        } else bits++;
    }
    if (bits) d[j] = uv;

    while (sz && !d[sz - 1]) sz--;
    if (sz <= 2 && v->u.bits.val != d) {
        memcpy(v->u.bits.val, d, sz);
        free(d);
        d = v->u.bits.val;
    }
    v->u.bits.data = d;
    v->u.bits.len = sz;
    v->u.bits.inv = 0;
    v->u.bits.bits = j * sizeof(bdigit_t) * 8 + bits;
    return l;
}

int bits_from_str(struct value_s *v, const struct value_s *v1) {
    int ch;

    if (actual_encoding) {
        uval_t uv = 0;
        unsigned int bits = 0;
        size_t j = 0, sz, osz;
        bdigit_t *d;
        struct value_s tmp;

        if (!v1->u.str.len) {
            v->obj = BITS_OBJ;
            v->u.bits.data = v->u.bits.val;
            v->u.bits.val[0] = 0;
            v->u.bits.len = 0;
            v->u.bits.inv = 0;
            v->u.bits.bits = 0;
            return 0;
        }

        if (v1->u.str.len <= 2 * sizeof(bdigit_t)) sz = 2;
        else {
            sz = v1->u.str.len / sizeof(bdigit_t);
            if (v1->u.str.len % sizeof(bdigit_t)) sz++;
        }
        d = bnew(&tmp, sz);

        encode_string(v1);
        while ((ch = encode_string(NULL)) != EOF) {
            uv |= (uint8_t)ch << bits;
            bits += 8;
            if (bits >= 8 * sizeof(bdigit_t)) {
                if (j >= sz) {
                    if (tmp.u.bits.val == d) {
                        sz = 16 / sizeof(bdigit_t);
                        d = (bdigit_t *)malloc(sz * sizeof(bdigit_t));
                        memcpy(d, tmp.u.bytes.val, j * sizeof(bdigit_t));
                    } else {
                        sz += 1024 / sizeof(bdigit_t);
                        if (sz < 1024 / sizeof(bdigit_t)) err_msg_out_of_memory(); /* overflow */
                        d = (bdigit_t *)realloc(d, sz * sizeof(bdigit_t));
                    }
                    if (!d) err_msg_out_of_memory();
                }
                d[j++] = uv;
                bits = uv = 0;
            }
        }
        if (bits) {
            if (j >= sz) {
                sz++;
                if (sz < 1) err_msg_out_of_memory(); /* overflow */
                if (tmp.u.bits.val == d) {
                    d = (bdigit_t *)malloc(sz * sizeof(bdigit_t));
                    memcpy(d, tmp.u.bytes.val, j * sizeof(bdigit_t));
                } else d = (bdigit_t *)realloc(d, sz * sizeof(bdigit_t));
                if (!d) err_msg_out_of_memory();
            }
            d[j++] = uv;
        }
        osz = j;

        while (osz && !d[osz - 1]) osz--;
        if (v == v1) v->obj->destroy(v);
        if (osz <= 2) {
            memcpy(v->u.bits.val, d, osz * sizeof(bdigit_t));
            if (tmp.u.bits.val != d) free(d);
            d = v->u.bits.val;
        } else if (osz < sz) {
            d = (bdigit_t *)realloc(d, osz * sizeof(bdigit_t));
            if (!d) err_msg_out_of_memory();
        }
        v->obj = BITS_OBJ;
        v->u.bits.data = d;
        v->u.bits.len = osz;
        v->u.bits.inv = 0;
        v->u.bits.bits = j * sizeof(bdigit_t) * 8 + bits;
        return 0;
    } 
    if (v1->u.str.chars == 1) {
        uint32_t ch2 = v1->u.str.data[0];
        if (ch2 & 0x80) utf8in(v1->u.str.data, &ch2);
        if (v == v1) v->obj->destroy(v);
        bits_from_u24(v, ch2);
        return 0;
    } 
    return 1;
}

void bits_from_bytes(struct value_s *v, const struct value_s *v1) {
    unsigned int bits = 0;
    size_t i = 0, j = 0, sz;
    bdigit_t *d, uv = 0;
    struct value_s tmp;

    if (!v1->u.bytes.len) {
        v->obj = BITS_OBJ;
        v->u.bits.data = v->u.bits.val;
        v->u.bits.val[0] = 0;
        v->u.bits.len = 0;
        v->u.bits.inv = 0;
        v->u.bits.bits = 0;
        return;
    }

    sz = v1->u.bytes.len / sizeof(bdigit_t);
    if (v1->u.bytes.len % sizeof(bdigit_t)) sz++;
    d = bnew(&tmp, sz);

    while (v1->u.bytes.len > i) {
        uv |= v1->u.bytes.data[i++] << bits;
        bits += 8;
        if (bits >= 8 * sizeof(bdigit_t)) {
            d[j++] = uv;
            bits = uv = 0;
        }
    }
    if (bits) {
        d[j] = uv;
        sz = j + 1;
    } else sz = j;

    while (sz && !d[sz - 1]) sz--;
    if (v == v1) v->obj->destroy(v);
    if (sz <= 2) {
        memcpy(v->u.bits.val, d, sz * sizeof(bdigit_t));
        if (tmp.u.bits.val != d) free(d);
        d = v->u.bits.val;
    }
    v->obj = BITS_OBJ;
    v->u.bits.data = d;
    v->u.bits.len = sz;
    v->u.bits.inv = 0;
    v->u.bits.bits = j * sizeof(bdigit_t) * 8 + bits;
}

static void calc1(oper_t op) {
    struct value_s *v1 = op->v1, *v = op->v;
    struct value_s tmp;
    uval_t uv;
    switch (op->op->u.oper.op) {
    case O_BANK:
        uv = v1->u.bits.len > 1 ? v1->u.bits.data[1] : 0;
        if (v1->u.bits.inv) uv = ~uv;
        if (v == v1) destroy(v);
        bits_from_u8(v, uv);
        return;
    case O_HIGHER:
        uv = v1->u.bits.len > 0 ? v1->u.bits.data[0] : 0;
        if (v1->u.bits.inv) uv = ~uv;
        if (v == v1) destroy(v);
        bits_from_u8(v, uv >> 8);
        return;
    case O_LOWER:
        uv = v1->u.bits.len > 0 ? v1->u.bits.data[0] : 0;
        if (v1->u.bits.inv) uv = ~uv;
        if (v == v1) destroy(v);
        bits_from_u8(v, uv);
        return;
    case O_HWORD:
        uv = v1->u.bits.len > 1 ? (v1->u.bits.data[1] << 16) : 0;
        uv |= v1->u.bits.len > 0 ? v1->u.bits.data[0] : 0;
        if (v1->u.bits.inv) uv = ~uv;
        if (v == v1) destroy(v);
        bits_from_u16(v, uv >> 8);
        return;
    case O_WORD:
        uv = v1->u.bits.len > 0 ? v1->u.bits.data[0] : 0;
        if (v1->u.bits.inv) uv = ~uv;
        if (v == v1) destroy(v);
        bits_from_u16(v, uv);
        return;
    case O_BSWORD:
        uv = v1->u.bits.len > 0 ? v1->u.bits.data[0] : 0;
        if (v1->u.bits.inv) uv = ~uv;
        if (v == v1) destroy(v);
        bits_from_u16(v, (uint8_t)(uv >> 8) | (uint16_t)(uv << 8));
        return;
    case O_INV:
        if (v != v1) copy(v1, v);
        v->u.bits.inv = !v->u.bits.inv;
        return;
    case O_NEG:
    case O_POS:
    case O_STRING:
        int_from_bits(&tmp, v1);
        if (v == v1) destroy(v);
        op->v1 = &tmp;
        tmp.refcount = 0;
        tmp.obj->calc1(op);
        op->v1 = v1;
        tmp.obj->destroy(&tmp);
        return;
    default:
        break;
    }
    obj_oper_error(op);
}

static void and_(const struct value_s *vv1, const struct value_s *vv2, struct value_s *vv) {
    size_t blen1, blen2;
    size_t i, len1, len2, sz;
    int neg1, neg2;
    bdigit_t *v1, *v2, *v;
    bdigit_t c;
    len1 = vv1->u.bits.len;
    len2 = vv2->u.bits.len;

    if (len1 <= 1 && len2 <= 1) {
        blen1 = vv1->u.bits.bits; blen2 = vv2->u.bits.bits;
        neg1 = vv1->u.bits.inv; neg2 = vv2->u.bits.inv;
        c = neg1 ? ~vv1->u.bits.val[0] : vv1->u.bits.val[0];
        c &= neg2 ? ~vv2->u.bits.val[0] : vv2->u.bits.val[0];
        neg1 &= neg2;
        if (neg1) c = ~c;
        v = vv->u.bits.val;
        vv->obj = BITS_OBJ; 
        vv->u.bits.data = v;
        vv->u.bits.bits = blen1 > blen2 ? blen1 : blen2;
        vv->u.bits.inv = neg1;
        v[0] = c;
        vv->u.bits.len = (c != 0);
        return;
    }
    if (len1 < len2) {
        const struct value_s *tmp = vv1; vv1 = vv2; vv2 = tmp;
        i = len1; len1 = len2; len2 = i;
    }
    v1 = vv1->u.bits.data; v2 = vv2->u.bits.data;
    blen1 = vv1->u.bits.bits; blen2 = vv2->u.bits.bits;
    neg1 = vv1->u.bits.inv; neg2 = vv2->u.bits.inv;
    if (blen1 < blen2) blen1 = blen2;

    sz = neg2 ? len1 : len2;
    v = bnew(vv, sz);

    if (neg1) {
        if (neg2) {
            for (i = 0; i < len2; i++) v[i] = v1[i] | v2[i];
            for (; i < len1; i++) v[i] = v1[i];
        } else {
            for (i = 0; i < len2; i++) v[i] = ~v1[i] & v2[i];
        }
    } else {
        if (neg2) {
            for (i = 0; i < len2; i++) v[i] = v1[i] & ~v2[i];
            for (; i < len1; i++) v[i] = v1[i];
        } else {
            for (i = 0; i < len2; i++) v[i] = v1[i] & v2[i];
        }
    }
    while (sz && !v[sz - 1]) sz--;
    if (sz <= 2 && v != vv->u.bits.val) {
        memcpy(vv->u.bits.val, v, sz * sizeof(bdigit_t));
        free(v);
        v = vv->u.bits.val;
    }
    if (vv == vv1 || vv == vv2) destroy(vv);
    vv->obj = BITS_OBJ; 
    vv->u.bits.data = v;
    vv->u.bits.len = sz;
    vv->u.bits.inv = neg1 & neg2;
    vv->u.bits.bits = blen1;
}

static void or_(const struct value_s *vv1, const struct value_s *vv2, struct value_s *vv) {
    size_t blen1, blen2;
    size_t i, len1, len2, sz;
    int neg1, neg2;
    bdigit_t *v1, *v2, *v;
    bdigit_t c;
    len1 = vv1->u.bits.len;
    len2 = vv2->u.bits.len;

    if (len1 <= 1 && len2 <= 1) {
        blen1 = vv1->u.bits.bits; blen2 = vv2->u.bits.bits;
        neg1 = vv1->u.bits.inv; neg2 = vv2->u.bits.inv;
        c = neg1 ? ~vv1->u.bits.val[0] : vv1->u.bits.val[0];
        c |= neg2 ? ~vv2->u.bits.val[0] : vv2->u.bits.val[0];
        neg1 |= neg2;
        if (neg1) c = ~c;
        v = vv->u.bits.val;
        vv->obj = BITS_OBJ; 
        vv->u.bits.data = v;
        vv->u.bits.bits = blen1 > blen2 ? blen1 : blen2;
        vv->u.bits.inv = neg1;
        v[0] = c;
        vv->u.bits.len = (c != 0);
        return;
    }
    if (len1 < len2) {
        const struct value_s *tmp = vv1; vv1 = vv2; vv2 = tmp;
        i = len1; len1 = len2; len2 = i;
    }
    v1 = vv1->u.bits.data; v2 = vv2->u.bits.data;
    blen1 = vv1->u.bits.bits; blen2 = vv2->u.bits.bits;
    neg1 = vv1->u.bits.inv; neg2 = vv2->u.bits.inv;
    if (blen1 < blen2) blen1 = blen2;

    sz = neg2 ? len2 : len1;
    v = bnew(vv, sz);

    if (neg1) {
        if (neg2) {
            for (i = 0; i < len2; i++) v[i] = v1[i] & v2[i];
        } else {
            for (i = 0; i < len2; i++) v[i] = v1[i] & ~v2[i];
            for (; i < len1; i++) v[i] = v1[i];
        }
    } else {
        if (neg2) {
            for (i = 0; i < len2; i++) v[i] = ~v1[i] & v2[i];
        } else {
            for (i = 0; i < len2; i++) v[i] = v1[i] | v2[i];
            for (; i < len1; i++) v[i] = v1[i];
        }
    }
    while (sz && !v[sz - 1]) sz--;
    if (sz <= 2 && vv->u.bits.val != v) {
        memcpy(vv->u.bits.val, v, sz * sizeof(bdigit_t));
        free(v);
        v = vv->u.bits.val;
    }
    if (vv == vv1 || vv == vv2) destroy(vv);
    vv->obj = BITS_OBJ; 
    vv->u.bits.data = v;
    vv->u.bits.len = sz;
    vv->u.bits.inv = neg1 | neg2;
    vv->u.bits.bits = blen1;
}

static void xor_(const struct value_s *vv1, const struct value_s *vv2, struct value_s *vv) {
    size_t blen1, blen2;
    size_t i, len1, len2, sz;
    int neg1, neg2;
    bdigit_t *v1, *v2, *v;
    bdigit_t c;
    len1 = vv1->u.bits.len;
    len2 = vv2->u.bits.len;

    if (len1 <= 1 && len2 <= 1) {
        blen1 = vv1->u.bits.bits; blen2 = vv2->u.bits.bits;
        neg1 = vv1->u.bits.inv; neg2 = vv2->u.bits.inv;
        c = vv1->u.bits.val[0] ^ vv2->u.bits.val[0];
        neg1 ^= neg2;
        v = vv->u.bits.val;
        vv->obj = BITS_OBJ; 
        vv->u.bits.data = v;
        vv->u.bits.bits = blen1 > blen2 ? blen1 : blen2;
        vv->u.bits.inv = neg1;
        v[0] = c;
        vv->u.bits.len = (c != 0);
        return;
    }
    if (len1 < len2) {
        const struct value_s *tmp = vv1; vv1 = vv2; vv2 = tmp;
        i = len1; len1 = len2; len2 = i;
    }
    v1 = vv1->u.bits.data; v2 = vv2->u.bits.data;
    blen1 = vv1->u.bits.bits; blen2 = vv2->u.bits.bits;
    neg1 = vv1->u.bits.inv; neg2 = vv2->u.bits.inv;
    if (blen1 < blen2) blen1 = blen2;

    sz = len1;
    v = bnew(vv, sz);

    for (i = 0; i < len2; i++) v[i] = v1[i] ^ v2[i];
    for (; i < len1; i++) v[i] = v1[i];
    while (sz && !v[sz - 1]) sz--;
    if (sz <= 2 && vv->u.bits.val != v) {
        memcpy(vv->u.bits.val, v, sz * sizeof(bdigit_t));
        free(v);
        v = vv->u.bits.val;
    }
    if (vv == vv1 || vv == vv2) destroy(vv);
    vv->obj = BITS_OBJ; 
    vv->u.bits.data = v;
    vv->u.bits.len = sz;
    vv->u.bits.inv = neg1 ^ neg2;
    vv->u.bits.bits = blen1;
}

static void concat(const struct value_s *vv1, const struct value_s *vv2, struct value_s *vv) {
    size_t blen = vv1->u.bits.bits + vv2->u.bits.bits;
    size_t sz, bits, i, j, rbits;
    bdigit_t *v1, *v, uv;
    bdigit_t inv = -vv2->u.bits.inv;
    struct value_s tmp;

    if (blen < vv2->u.bits.bits) err_msg_out_of_memory(); /* overflow */
    sz = blen / 8 / sizeof(bdigit_t);
    if (blen % (8 * sizeof(bdigit_t))) sz++;
    v = bnew(&tmp, sz);

    v1 = vv2->u.bits.data;
    rbits = vv2->u.bits.bits / 8 / sizeof(bdigit_t);
    bits = vv2->u.bits.bits & (8 * sizeof(bdigit_t) - 1);
    for (i = 0; i < rbits && i < vv2->u.bits.len; i++) v[i] = v1[i] ^ inv;
    for (; i < rbits; i++) v[i] = inv;
    if (i < vv2->u.bits.len) uv = v1[i] ^ inv; else uv = inv;
    if (inv) uv &= (1 << bits) - 1;
    inv = vv1->u.bits.inv;

    rbits = vv1->u.bits.bits / 8 / sizeof(bdigit_t);
    v1 = vv1->u.bits.data;

    if (bits) {
        for (j = 0; j < vv1->u.bits.len; j++) {
            v[i++] = uv | (v1[j] << bits);
            uv = v1[j] >> (8 * sizeof(bdigit_t) - bits);
        }
        if (j < rbits) { v[i++] = uv; uv = 0; j++;}
        for (; j < rbits; j++) v[i++] = 0;
        if (i < sz) v[i] = uv;
    } else {
        for (j = 0; j < vv1->u.bits.len; j++) v[i++] = v1[j];
        for (; j < rbits; j++) v[i++] = 0;
    }

    while (sz && !v[sz - 1]) sz--;
    if (vv == vv1 || vv == vv2) vv->obj->destroy(vv);
    if (sz <= 2) {
        memcpy(vv->u.bits.val, v, sz * sizeof(bdigit_t));
        if (tmp.u.bits.val != v) free(v);
        v = vv->u.bits.val;
    }
    vv->obj = BITS_OBJ; 
    vv->u.bits.data = v;
    vv->u.bits.len = sz;
    vv->u.bits.bits = blen;
    vv->u.bits.inv = inv;
}

static void lshift(const struct value_s *vv1, const struct value_s *vv2, size_t s, struct value_s *vv) {
    size_t i, sz, bits, len1, word;
    int bit, inv;
    bdigit_t *v1, *v, *o;

    word = s / 8 / sizeof(bdigit_t);
    bit = s % (8 * sizeof(bdigit_t));
    v1 = vv1->u.bits.data;
    len1 = vv1->u.bits.len;
    bits = vv1->u.bits.bits + s;
    if (bits < s) err_msg_out_of_memory(); /* overflow */
    inv = vv1->u.bits.inv;
    sz = word + (bit > 0);
    sz += len1;
    if (sz < len1) err_msg_out_of_memory(); /* overflow */
    v = bnew(vv, sz);
    o = &v[word];
    if (bit) {
        o[len1] = 0;
        for (i = len1; i--;) {
            o[i + 1] |= v1[i] >> (8 * sizeof(bdigit_t) - bit);
            o[i] = v1[i] << bit;
        }
    } else if (len1) memmove(o, v1, len1 * sizeof(bdigit_t));
    memset(v, 0, word * sizeof(bdigit_t));

    while (sz && !v[sz - 1]) sz--;
    if (sz <= 2 && v != vv->u.bits.val) {
        memcpy(vv->u.bits.val, v, sz * sizeof(bdigit_t));
        free(v);
        v = vv->u.bits.val;
    }
    if (vv == vv1 || vv == vv2) vv->obj->destroy(vv);
    vv->obj = BITS_OBJ;
    vv->u.bits.data = v;
    vv->u.bits.len = sz;
    vv->u.bits.bits = bits;
    vv->u.bits.inv = inv;
}

static void rshift(const struct value_s *vv1, const struct value_s *vv2, uval_t s, struct value_s *vv) {
    size_t i, sz, bits, word;
    int bit, inv;
    bdigit_t *v1, *v;

    word = s / 8 / sizeof(bdigit_t);
    bit = s - word * 8 * sizeof(bdigit_t);
    bits = vv1->u.bits.bits;
    sz = vv1->u.bits.len - word;
    if (sz <= 0 || bits <= s) {
        if (vv == vv1 || vv == vv2) vv->obj->destroy(vv);
        vv->obj = BITS_OBJ;
        vv->u.bits.data = vv->u.bits.val;
        vv->u.bits.val[0] = 0;
        vv->u.bits.len = 0;
        vv->u.bits.inv = 0;
        vv->u.bits.bits = 0;
        return;
    }
    bits -= s;
    v = bnew(vv, sz);
    v1 = vv1->u.bits.data + word;
    if (bit) {
        for (i = 0; i < sz - 1; i++) {
            v[i] = v1[i] >> bit;
            v[i] |= v1[i + 1] << (8 * sizeof(bdigit_t) - bit);
        }
        v[i] = v1[i] >> bit;
    } else if (sz) memmove(v, v1, sz * sizeof(bdigit_t));

    while (sz && !v[sz - 1]) sz--;
    if (sz <= 2 && v != vv->u.bits.val) {
        memcpy(vv->u.bits.val, v, sz * sizeof(bdigit_t));
        free(v);
        v = vv->u.bits.val;
    }
    inv = vv1->u.bits.inv;
    if (vv == vv1 || vv == vv2) vv->obj->destroy(vv);
    vv->obj = BITS_OBJ;
    vv->u.bits.data = v;
    vv->u.bits.len = sz;
    vv->u.bits.bits = bits;
    vv->u.bits.inv = inv;
}

static void repeat(oper_t op, uval_t rep) {
    struct value_s *vv1 = op->v1, *vv = op->v;
    bdigit_t *v, *v1;
    bdigit_t uv;
    size_t sz, i, j, rbits, abits, bits;
    size_t blen = vv1->u.bits.bits;

    if (!rep || !blen) {
        if (vv1 == vv) destroy(vv);
        copy(&null_bits, vv); return;
    }
    if (rep == 1) {
        if (vv1 != vv) copy(vv1, vv);
        return;
    }

    if (blen > SIZE_MAX / rep) err_msg_out_of_memory(); /* overflow */
    blen *= rep;
    sz = blen / 8 / sizeof(bdigit_t);
    if (blen % (8 * sizeof(bdigit_t))) sz++;

    v = bnew(vv, sz);
    v1 = vv1->u.bits.data;

    i = 0;
    bits = 0;
    uv = 0;
    rbits = vv1->u.bits.bits / 8 / sizeof(bdigit_t);
    abits = vv1->u.bits.bits % (8 * sizeof(bdigit_t));
    while (rep--) {
        if (bits) {
            for (j = 0; j < rbits && j < vv1->u.bits.len; j++) {
                v[i++] = uv | (v1[j] << bits);
                uv = (v1[j] >> (8 * sizeof(bdigit_t) - bits));
            }
            uv |= v1[j] << bits;
            if (j < rbits) { v[i++] = uv; uv = 0; j++;}
            for (; j < rbits; j++) v[i++] = 0;
            bits += abits;
            if (bits >= 8 * sizeof(bdigit_t)) {
                v[i++] = uv;
                bits -= 8 * sizeof(bdigit_t);
                if (bits) uv = v1[j] >> (abits - bits);
                else uv = 0;
            }
        } else {
            for (j = 0; j < rbits && j < vv1->u.bits.len; j++) v[i++] = v1[j];
            for (; j < rbits; j++) v[i++] = 0;
            uv = v1[j];
            bits = abits;
        }
    }
    if (i < sz) v[i] = uv;

    while (sz && !v[sz - 1]) sz--;
    if (sz <= 2 && vv->u.bits.val != v) {
        memcpy(vv->u.bits.val, v, sz * sizeof(bdigit_t));
        free(v);
        v = vv->u.bits.val;
    }
    if (vv == vv1 || vv == op->v2) destroy(vv);
    vv->obj = BITS_OBJ; 
    vv->u.bits.data = v;
    vv->u.bits.len = sz;
    vv->u.bits.bits = blen;
}

static MUST_CHECK struct value_s *calc2(oper_t op) {
    struct value_s *v1 = op->v1, *v2 = op->v2, *v = op->v;
    struct value_s tmp, *result;
    ival_t shift;
    switch (v2->obj->type) {
    case T_BOOL:
        bits_from_bool(&tmp, op->v2->u.boolean);
        //if (v2 == v) v->obj->destroy(v);
        op->v2 = &tmp;
        tmp.refcount = 0;
        result = calc2(op);
        op->v2 = v2;
        return result;
    case T_BITS: 
        switch (op->op->u.oper.op) {
        case O_AND: and_(v1, v2, v); return NULL;
        case O_OR: or_(v1, v2, v); return NULL;
        case O_XOR: xor_(v1, v2, v); return NULL;
        case O_CONCAT: concat(v1, v2, v); return NULL;
        default: break;
        }
        /* fall through */
    case T_INT:
        switch (op->op->u.oper.op) {
        case O_LSHIFT:
            if (v2->obj->ival(v2, &tmp, &shift, 8*sizeof(ival_t), &op->epoint2)) {
                if (v1 == v || v2 == v) v->obj->destroy(v);
                tmp.obj->copy_temp(&tmp, v);
                return NULL;
            }
            if (shift < 0) rshift(v1, v2, -shift, v); else lshift(v1, v2, shift, v);
            return NULL;
        case O_RSHIFT:
            if (v2->obj->ival(v2, &tmp, &shift, 8*sizeof(ival_t), &op->epoint2)) {
                if (v1 == v || v2 == v) v->obj->destroy(v);
                tmp.obj->copy_temp(&tmp, v);
                return NULL;
            }
            if (shift < 0) lshift(v1, v2, -shift, v); else rshift(v1, v2, shift, v);
            return NULL;
        default: break;
        }
        int_from_bits(&tmp, v1);
        if (v1 == v) {v->obj->destroy(v); v->obj = NONE_OBJ;}
        op->v1 = &tmp;
        tmp.refcount = 0;
        result = tmp.obj->calc2(op);
        op->v1 = v1;
        tmp.obj->destroy(&tmp);
        return result;
    default: 
        if (op->op != &o_MEMBER) {
            return op->v2->obj->rcalc2(op);
        }
    }
    obj_oper_error(op);
    return NULL;
}

static MUST_CHECK struct value_s *rcalc2(oper_t op) {
    struct value_s *v1 = op->v1, *v2 = op->v2, *v = op->v;
    struct value_s tmp, *result;
    switch (v1->obj->type) {
    case T_BOOL:
        bits_from_bool(&tmp, op->v1->u.boolean);
        //if (v1 == v) v->obj->destroy(v);
        op->v1 = &tmp;
        tmp.refcount = 0;
        result = rcalc2(op);
        op->v1 = v1;
        return result;
    case T_BITS: 
        switch (op->op->u.oper.op) {
        case O_AND: and_(v1, v2, v); return NULL;
        case O_OR: or_(v1, v2, v); return NULL;
        case O_XOR: xor_(v1, v2, v); return NULL;
        case O_CONCAT: concat(v1, v2, v); return NULL;
        default: break;
        }
        /* fall through */
    case T_INT:
        int_from_bits(&tmp, v2);
        if (v2 == v) {v->obj->destroy(v); v->obj = NONE_OBJ;}
        op->v2 = &tmp;
        tmp.refcount = 0;
        result = tmp.obj->rcalc2(op);
        op->v2 = v2;
        tmp.obj->destroy(&tmp);
        return result;
    default:
        if (op->op != &o_IN) {
            return v1->obj->calc2(op);
        }
    }
    obj_oper_error(op); 
    return NULL;
}

static inline MUST_CHECK struct value_s *slice(struct value_s *vv1, uval_t ln, ival_t offs, ival_t end, ival_t step, struct value_s *vv) {
    size_t bo, wo, bl, wl, sz;
    bdigit_t uv;
    bdigit_t *v, *v1;
    bdigit_t inv = -vv1->u.bits.inv;
    int bits;
    struct value_s tmp;

    if (!ln) {
        if (vv1 == vv) destroy(vv);
        copy(&null_bits, vv);return NULL;
    }
    if (step == 1) {
        if (ln == vv1->u.bits.len) {
            if (vv1 != vv) copy(vv1, vv);
            return NULL; /* original bits */
        }

        bo = offs % (8 * sizeof(bdigit_t));
        wo = offs / (8 * sizeof(bdigit_t));
        bl = ln % (8 * sizeof(bdigit_t));
        wl = ln / (8 * sizeof(bdigit_t));

        sz = wl + (bl > 0);
        v = bnew(&tmp, sz);

        v1 = vv1->u.bits.data + wo;
        if (bo) {
            for (sz = 0; sz < wl; sz++) {
                v[sz] = inv ^ (v1[sz] >> bo) ^ (v1[sz + 1] << (8 * sizeof(bdigit_t) - bo));
            }
            if (bl) {
                v[sz] = v1[sz] >> bo;
                if (bl > (8 * sizeof(bdigit_t) - bo)) v[sz] |= v1[sz + 1] << (8 * sizeof(bdigit_t) - bo);
                v[sz] ^= inv;
            }
        } else {
            for (sz = 0; sz < wl; sz++) v[sz] = v1[sz] ^ inv;
            if (bl) v[sz] = v1[sz] ^ inv;
        }
        if (bl) v[sz++] &= ((1 << bl) - 1);
    } else {
        sz = ln / 8 / sizeof(bdigit_t);
        if (ln % (8 * sizeof(bdigit_t))) sz++;
        v = bnew(&tmp, sz);

        uv = inv;
        sz = bits = 0;
        while ((end > offs && step > 0) || (end < offs && step < 0)) {
            wo = offs / 8 / sizeof(bdigit_t);
            if (wo < vv1->u.bits.len && ((vv1->u.bits.data[wo] >> (offs & (8 * sizeof(bdigit_t) - 1))) & 1)) {
                uv ^= 1 << bits;
            }
            bits++;
            if (bits == 8 * sizeof(bdigit_t)) {
                v[sz++] = uv;
                uv = inv;
                bits = 0;
            }
            offs += step;
        }
        if (bits) v[sz++] = uv & ((1 << bits) - 1);
    }
    while (sz && !v[sz - 1]) sz--;
    if (vv == vv1) vv->obj->destroy(vv);
    if (sz <= 2) {
        memcpy(vv->u.bits.val, v, sz * sizeof(bdigit_t));
        if (tmp.u.bits.val != v) free(v);
        v = vv->u.bits.val;
    }
    vv->obj = BITS_OBJ;
    vv->u.bits.bits = ln;
    vv->u.bits.len = sz;
    vv->u.bits.inv = 0;
    vv->u.bits.data = v;
    return NULL;
}

static MUST_CHECK struct value_s *iindex(oper_t op) {
    size_t ln, sz;
    ival_t offs;
    size_t i, o;
    struct value_s *vv1 = op->v1, *vv2 = op->v2, *vv = op->v;
    bdigit_t *v;
    bdigit_t uv;
    bdigit_t inv = -vv1->u.bits.inv;
    int bits;
    struct value_s tmp, err;

    ln = vv1->u.bits.bits;

    if (vv2->obj == LIST_OBJ) {
        if (!vv2->u.list.len) {
            if (vv1 == vv) destroy(vv);
            copy(&null_bits, vv);return NULL;
        }
        sz = (vv2->u.list.len + 8 * sizeof(bdigit_t) - 1) / (8 * sizeof(bdigit_t));

        v = bnew(&tmp, sz);

        uv = inv;
        bits = sz = 0;
        for (i = 0; i < vv2->u.list.len; i++) {
            offs = indexoffs(vv2->u.list.data[i], &err, ln, &op->epoint2);
            if (offs < 0) {
                if (tmp.u.bits.val != v) free(v);
                if (vv1 == vv) destroy(vv);
                err.obj->copy_temp(&err, vv);
                return NULL;
            }
            o = offs / 8 / sizeof(bdigit_t);
            if (o < vv1->u.bits.len && ((vv1->u.bits.data[o] >> (offs & (8 * sizeof(bdigit_t) - 1))) & 1)) {
                uv ^= 1 << bits;
            }
            bits++;
            if (bits == 8 * sizeof(bdigit_t)) {
                v[sz++] = uv;
                uv = inv;
                bits = 0;
            }
        }
        if (bits) v[sz++] = uv & ((1 << bits) - 1);

        while (sz && !v[sz - 1]) sz--;
        if (vv == vv1 || vv == vv2) vv->obj->destroy(vv);
        if (sz <= 2) {
            memcpy(vv->u.bits.val, v, sz * sizeof(bdigit_t));
            if (tmp.u.bits.val != v) free(v);
            v = vv->u.bits.val;
        }
        vv->obj = BITS_OBJ;
        vv->u.bits.bits = vv2->u.list.len;
        vv->u.bits.len = sz;
        vv->u.bits.inv = 0;
        vv->u.bits.data = v;
        return NULL;
    }
    if (vv2->obj == COLONLIST_OBJ) {
        ival_t length, end, step;
        length = sliceparams(op, ln, &offs, &end, &step);
        if (length < 0) return NULL;
        return slice(vv1, length, offs, end, step, vv);
    }
    offs = indexoffs(vv2, &err, ln, &op->epoint2);
    if (offs < 0) {
        if (vv1 == vv) destroy(vv);
        err.obj->copy_temp(&err, vv);
        return NULL;
    }

    uv = inv;
    o = offs / 8 / sizeof(bdigit_t);
    if (o < vv1->u.bits.len && ((vv1->u.bits.data[o] >> (offs & (8 * sizeof(bdigit_t) - 1))) & 1)) {
        uv ^= 1;
    }
    if (vv == vv1 || vv == vv2) vv->obj->destroy(vv);
    bits_from_bool(vv, uv & 1);
    return NULL;
}

void bitsobj_init(void) {
    obj_init(&obj, T_BITS, "<bits>");
    obj.destroy = destroy;
    obj.copy = copy;
    obj.copy_temp = copy_temp;
    obj.same = same;
    obj.truth = truth;
    obj.hash = hash;
    obj.repr = repr;
    obj.ival = ival;
    obj.uval = uval;
    obj.real = real;
    obj.sign = sign;
    obj.abs = absolute;
    obj.integer = integer;
    obj.len = len;
    obj.calc1 = calc1;
    obj.calc2 = calc2;
    obj.rcalc2 = rcalc2;
    obj.repeat = repeat;
    obj.iindex = iindex;
}
